;*****************************************************************************
;*****************************************************************************
;
; Module Name:	QUICC.ASM
; Description:	Intelligent Display Manager (IDM) code for PIC16C84/PIC16F84
; System:	QUICC-IDM Controller Board
;
; Original Author:	RAC
; Original Date:	09.02.97
;
; PIC Configuration:
;	Device: PIC16C84 or PIC16F84
;	Oscillator: 4MHz XTAL
;	Watcdog: Enabled
;	Power up timer: Enabled
;
; Modification Record:
;	09.02.97 0.1 	RAC	Initial version
;	22.06.97 1.0 	RAC	First released version
;	13.07.97 1.1 	RAC	Added new display reset command
;				Fixed bug in Wdog restart code
;				Made boot_portB reliably obtain panel data
;	09.08.97 1.2 	RAC	Fixed interrupt FSR corruption bug
;	10.09.97 1.3 	RAC	Refinements and code flattening
;	17.10.97 1.4 	RAC	New boot hold mode and standby disable
;	21.12.97 1.5 	RAC	Fixed boot_portB bug in boot hold mode
;	04.02.98 1.6 	RAC	Added "lamp test" facilities
;
;
;*****************************************************************************
;*****************************************************************************

    ifdef CTYPE
	processor 16c84 
	__config B'11111111111101'
	include "p16c84.inc"
    else
	processor 16f84 
	__config B'11111111110101'
	include "p16f84.inc"
    endif
  
;*****************************************************************************
;********************* CONSTANTS AND DATA DECLARATIONS ***********************
;*****************************************************************************

;-----------------------------------------------------------------------------
; Firmware revision number - sent to PC (must not be zero)

; NOTE: Top nybble is major version, and bottom nybble is minor version

firmware_ver	equ	16h

;*****************************************************************************
; I/O port allocations

;-----------------------------------------------------------------------------
; PORT A allocations

; NOTE: To avoid potential problems with read-modify-write instructions,
; the code should avoid performing multiple BSF/BCF instructions to different
; bits on the same port without some intervening instruction. This is purely
; a precautionary measure as no problems have been experienced in practice.

FLASH1	equ	0	;flash clock 1 output
FLASH2	equ	1	;flash clock 2 output
DP_RST	equ	2	;display panel reset (active low)
PSU_ON	equ	3	;PSU control (active high)
FAIL_LD	equ	4	;Failure LED (active low)

;-----------------------------------------------------------------------------
; PORT B allocations

RX_PC	equ	0	;receive line from PC (PC transmit - INT0)
WDOG1	equ	1	;watchdog from panel 1 (via switch - fail low)
WDOG2	equ	2	;watchdog from panel 2 (via switch - fail low)
WDOG3	equ	3	;watchdog from panel 3 (via switch - fail low)
SBY_SW	equ	4	;standby enable switch (active low)
PC_DTR	equ	5	;DTR signal from PC (flash clock + sby watchdog)
TX_PC	equ	6	;transmit line to PC (PC receive)
BOOT_LD	equ	7	;boot/status LED (active low)

;-----------------------------------------------------------------------------
; Port directions and initialisation values

PB_TRIS	equ	B'00111111'     ; port B all inputs except RB6 amd RB7
PA_TRIS	equ	B'11100000'     ; port A all outputs
OPT_VAL equ	B'00001110'     ; PRE-SCALAR WDT, INT0 H->L, pullups active

PB_DFLT	equ	B'11000000'	; default port B value (PC TX set high)
PA_DFLT	equ	B'00010000'	; default port A value (panels IN reset)

;*****************************************************************************
; RAM data allocations (0ch to 2fh are valid locations)

;-----------------------------------------------------------------------------
; miscellaneous register stores

W_TEMP		equ	0ch	; store for W register
S_TEMP		equ	0dh	; store for status register
FS_TEMP		equ	0eh	; store for FSR register

CT_INT1		equ	0fh	; counter 1 for interrupt routine
CT_INT2		equ	10h	; counter 2 for interrupt routine
CT_SBY		equ	11h	; standby counter (dec by intr, reset by main)

COUNT1   	equ     12h	; general purpose counter 1
COUNT2   	equ     13h	; general purpose counter 2

M_FLAGS		equ	14h	; miscellaneous flags store
S_FLAGS		equ	15h	; status flags store

;-----------------------------------------------------------------------------
; stores specific to RS232 comms

SerialStatus	equ	16h	;serial status register
BitCount	equ	17h	;bitcount for transmission/reception
HoldReg		equ	18h	;transmitter/receiver holding register
RxStore		equ	19h	;receiver output data store

;-----------------------------------------------------------------------------
;high level status stores

comms_mode	equ	1ah	;serial comms mode
cmd_store	equ	1bh	;incoming command store
data_store	equ	1ch	;incoming data store
csum_store	equ	1dh	;checksum store (TX and RX comms)

init_portb	equ	1eh	;initial port B status on startup
trip_portb	equ	1fh	;last port B status on watchdog trip
boot_portb	equ	20h	;last port B status during boot (psu off)

sby_tmr		equ	21h	;store for standby timer
boot_tmr	equ	22h	;store for boot timer
psu_tmr		equ	23h	;store for PSU timer
csum_eeprom	equ	24h	;EEPROM checksum store

;-----------------------------------------------------------------------------
;flash clock stores

flash1_set	equ	25h	;store for flash clock 1 settings
flash2_set	equ	26h	;store for flash clock 2 settings

flash1_cnt	equ	27h	;store for flash clock 1 counters
flash2_cnt	equ	28h	;store for flash clock 2 counters

;-----------------------------------------------------------------------------
; circular receive buffer and definitions

;IMPORTANT! Memory locations 2bh to 2fh inclusive are used by the buffer,
;and should therefore be avoided by anything else. The RxBuffEnd location
;marks the location AFTER the last one used, and actually points at void 
;space. Note that the number of characters that can be stored is
;ONE LESS than the buffer size, since the head and tail pointers need to
;be be different to indicate characters present in the buffer.

#define rx_buff_size	5	;length of receive buffer

RxHead		equ	29h	;buffer head pointer
RxTail		equ	2ah	;buffer tail pointer

RxBuff		equ	2bh
RxBuffEnd	equ	(RxBuff + rx_buff_size)

;*****************************************************************************
; Varous constants and definitions for this program

;-----------------------------------------------------------------------------
; miscellaneous constants

TRUE	equ	1		; generic TRUE condition
FALSE	equ	0		; generic FALSE condition

;-----------------------------------------------------------------------------
; INTERNAL short timing definitions (multiples of timer overflow int)

; The interrupt timers are fully under control of the interrupt routine, and
; apart from the initial load, should NOT be altered by the main code.
; Note that these timers will NOT be adjusted when serial comms is active.

TM_INTR1	equ	.250	; 250 * 256us = 64ms seconds fast int tick
TM_INTR2	equ	.8	; 8 * 64ms = 0.512 seconds slow int tick

;-----------------------------------------------------------------------------
; INTERNAL short timing definitions - for delay calculations

; IMPORTANT! Since the RTCC is used by the serial comms, and this affects the
; intervals at which it overflows, any delay calculation must take this into
; account. Therefore the timer delay system checks the comms mode, and 
; decrements its counter by a different amount, depending whether comms is
; active or otherwise. The resulting delay is usable, but not highly accurate.

TL_CNT1		equ	.255	;20ms (approx) counter 1

TDEC_IDLE	equ	.3	;counter 1 decrement when idle
TDEC_COMMS	equ	.1	;counter 1 decrement when comms active

;-----------------------------------------------------------------------------
; standby timer definitions (multiples of interrupt counter TM_INTR2)

; IMPORTANT! Due to the limited instruction set, the standby timer works in 
; a rather unusual way: the interrupt routine decrements it, unless to do so 
; would make it zero. To check whether it's expired, the main code also 
; checks to see if decrementing it would make it zero, whereupon it enters 
; standby mode. Whenever a DTR toggle is detected, the counter is reset to 
; maximum. Note that when serial comms is being progressed, the standby
; timer does not decrement; this is acceptable, as the standby timer should
; only activate if the PC stops all input activity. Note that the actual value
; should be 2 greater than the desired time, to compensate for the way
; the system operates.

TS_SBY		equ	.84	; (84 - 2) * 0.51s = 42 seconds standby timer

;-----------------------------------------------------------------------------
; timing delays - in multiples of 20ms (max time 5.1 seconds)

; These timers are used by the main loop code, and are passed to the
; wait_delay routine. Though the delays are not highly accurate, they will
; operate whether serial comms is being received or not.

TM_BOOT		equ	.20	; 20 * 20mS = 0.4 seconds dflt boot time
TM_PSU		equ	.15	; 15 * 20mS = 0.3 seconds dflt PSU on time

TM_START	equ	.10	; 10 * 20mS = 0.2 seconds startup holdoff

; The initial port B read delay specifies how long the system waits before
; sampling the port B status after putting the panels in reset. The boot 
; port B specifies how long the system waits before sampling the boot port B
; status after taking the panels out of reset. If this is set shorter than
; the panel recovery time (see below), the boot status will read correctly,
; even if the PSU is outputting voltage at the time. 

TM_INITB	equ	.1	; 1 * 20mS = 0.02 seconds init portB read
TM_BOOTB	equ	.1	; 1 * 20mS = 0.02 seconds boot portB read

TM_FAIL1	equ	.5	; 5 * 20mS = 0.1 seconds failure register
TM_FLASH	equ	.15	; 15 * 20mS = 0.3 seconds fail LED flash
TM_IFAULT	equ	.50	; 50 * 20mS = 1 seconds fail LED illuminate

TM_WHOLD	equ	.255	; 255 * 20mS = 5.1 seconds wdog trip holdoff

; The display reset time is a nominal value chosen to ensure the displays
; are actually reset. The reset recovery time is more important, chosen to
; ensure that the panels have started scanning again before the watchdogs
; are monitored. With the old display firmware, the displays took between 
; 60ms and 80ms to recover; this may be longer with the enhanced firmware,
; so the timer is set to a higher value, 240ms. Note that the timer should
; not be excessively long, as this could put the system at risk under fault
; situations.

TM_DRESET	equ	.5	; 5 * 20mS = 0.1 seconds disp panel reset
TM_RRECOV	equ	.12	; 12 * 20mS = 0.24 seconds reset recovery

;-----------------------------------------------------------------------------
;config switch specifiers

#define _cfg_sw1	init_portb, WDOG1	;config switch 1
#define _cfg_sw2	init_portb, WDOG2	;config switch 2
#define _cfg_sw3	init_portb, WDOG3	;config switch 3

;-----------------------------------------------------------------------------
;miscellaneous flags (M_FLAGS) store bits (remember top bit not avail to PC)

#define	_tim_overflow	M_FLAGS, 0	;interrupt timer overflow bit
#define	_dtr_status	M_FLAGS, 1	;DTR status store
#define	_sby_request	M_FLAGS, 2	;request standby mode
#define	_start_request	M_FLAGS, 3	;request system startup
#define	_recover_req	M_FLAGS, 4	;request recovery from fault mode
#define	_dreset_req	M_FLAGS, 5	;request display panel reset
#define	_boot_hold_req	M_FLAGS, 6	;request system boot hold
#define	_flamp_test_req	M_FLAGS, 7	;request "fault" lamp test

;-----------------------------------------------------------------------------
;status flags (S_FLAGS) store bits (remember top bit not avail to PC)

#define	_sby_status	S_FLAGS, 0	;standby status flag
#define	_fault_status	S_FLAGS, 1	;fault status flag
#define	_sys_started	S_FLAGS, 2	;system started from reset
#define	_pic_wdog_trip	S_FLAGS, 3	;reset due to PIC wdog trip
#define	_pic_wdog_boot	S_FLAGS, 4	;pending watchdog re-boot
#define	_panel_boot	S_FLAGS, 5	;normal panel boot in progress
#define	_bad_ee_check	S_FLAGS, 6	;EEPROM checksum bad

;-----------------------------------------------------------------------------
;comms mode definitions - see handle_comms function for descriptions

cm_wait_zero	equ	.0	;waiting for 0
cm_wait_head	equ	.1	;waiting for startup header
cm_wait_cmd1	equ	.2	;waiting for command code
cm_wait_data	equ	.3	;waiting for data val
cm_wait_csum	equ	.4	;waiting for checksum
cm_wait_term	equ	.5	;waiting for terminator

;-----------------------------------------------------------------------------
;comms header and mask values

quicc_idm_hdr	equ	0feh	;header code for QUICC-IDM
quicc_idm_term	equ	0efh	;terminator code for QUICC-IDM

stat_portb_mask	equ	040h	;mask to be ORed with portB status data
stat_sflag_mask	equ	080h	;mask to be ORed with status data

;-----------------------------------------------------------------------------
;command code values

cmdc_get_stat	equ	.40	;simply request status
cmdc_sys_start	equ	.41	;wake up system
cmdc_go_sleep	equ	.42	;put system to sleep
cmdc_boot_hold	equ	.43	;out of reset, but PSU off (boot LED on)

cmdc_tmr_dflt	equ	.50	;load default values into all timers
cmdc_set_sby	equ	.51	;set standby timer value (mult of 0.51 secs)
cmdc_set_boot	equ	.52	;set boot timer value (mult of 0.02 secs)
cmdc_set_psu	equ	.53	;set psu ovr timer value (mult of 0.02 secs)
cmdc_epm_load	equ	.54	;reload all timer values from EEPROM
cmdc_enab_eewr	equ	.55	;enable EEPROM writes

cmdc_reboot	equ	.60	;allow system to re-boot (PIC watchdog trip)
cmdc_recover	equ	.61	;recover from fault reporting mode
cmdc_dreset	equ	.62	;reset display panels (run or fault mode)
cmdc_flamp_test	equ	.63	;illuminate the "fault" LED briefly

cmdc_flash_rst	equ	.70	;reset flash 1 and flash 2 clocks
cmdc_flash1_set	equ	.71	;set flash clock 1 division ratios
cmdc_flash2_set	equ	.72	;set flash clock 2 division ratios
cmdc_flash_inv	equ	.73	;inverse reset flash 1 and flash 2 clocks
cmdc_flash1_cnt	equ	.74	;set flash clock 1 counter value
cmdc_flash2_cnt	equ	.75	;set flash clock 2 counter value

;-----------------------------------------------------------------------------
; definitions specific to RS232 comms

#define	_txmtProgress	SerialStatus,0	;transmission in progress
#define	_txmtStart	SerialStatus,1	;ready for start bit transmission
#define	_txmtRequest	SerialStatus,2	;transmission request -lock receive

#define	_rcvProgress	SerialStatus,3	;reception in progress
#define	_rcvTemp	SerialStatus,4	;temp bit store for receiver
#define	_rcvOver	SerialStatus,5	;reception finished - char available
#define	_FrameErr	SerialStatus,6	;framing error

#define	TX_PIN	PORTB,TX_PC	; TX to PC on port B
#define	RX_PIN	PORTB,RX_PC	; RX bit on port B

;-----------------------------------------------------------------------------
;bit count and other misc values

TX_BITC_LOAD	equ	.9	; 8 data + 1 stop (start not included)
RX_BITC_LOAD	equ	.9	; 8 data + 1 stop (start not included)

;-----------------------------------------------------------------------------
; preload values for RTCC (including fiddle factors)

;NOTE: At 12800 baud, theoretical interval is 78 for 1 bit (117 for 1.5 bits)
;The values below are adjusted by "fiddle factors" to compensate for the
;very significant code overhead presented by the start of the interrupt
;routine. If any code changes are made, these will most definitely NEED
;to be recalculated. Remember that RTCC counts UPWARDS, so you need to take
;the required interval from 256 to get the correct load value.

;Remember that:
;
;1) There are 3 to 4 cycles latency from start bit detection to the interrupt
;   routine being executed.
;2) There are 2 cycles latency between writing to the timer and it starting
;   to count. Also, there are 4 cycles latency from the timer hitting 255 and
;   the timer interrupt routine being called.
;3) Each bit is read slightly later than the timer reload interval. This will
;   not affect the re-load value, but will affect the start bit timing value,
;   which should go to the bit read, NOT the reload.
;4) By the time the start bit routine loads the counter, several instruction
;   cycles will have passed, but not enough to compromise the start bit
;   read reliability. Current code will be about 20% into the start bit time.
;5) The reload counter is FAR more critical than the start bit counter, as
;   errors will be multiplied up by a factor of 10! A single instruction
;   cycle error will result in 10us of slippage, or 13% of one bit interval.
;   If the start bit counter is inaccurate, the data reads will occur later
;   or earlier in ALL bits, aggravating any error in the reload interval.
;6) The end of transmission/reception load value should ideally be zero, so 
;   that a maximum interval elapses to the next character. This will keep the
;   timer calculation routines as accurate as possible.
;7) One unavoidable problem is if the start bit fires off when a timer 
;   interrupt is currently being serviced. This will delay reading of all 
;   further bit positions. For this reason, it is vital to keep the ISR
;   as short as possible, so that all bit reads will still occur under
;   worst case conditions.
;8) To help compensate a little for the ISR firing off, a small reduction
;   to the start timer is made. This means bit reads will be slightly earlier,
;   thereby allowing for more slippage on the start bit detection.
;9) The INT0 interrupt is permanently enabled, in order to keep the ISR as
;   small and clean as possible. This means that interrupts will fire off 
;   during reception, but will not interfere with the timing, bevause they
;   are in antiphase to the timer interrupts. However, this is another good
;   reason for keeping the ISR as short as possible.

;Also note that when neither transmission nor reception are taking place, the
;RTCC will idle at the maximum interval (in order to reduce processor 
;overhead). This must be taken into account by any timing delay routines.

RTCC_TX_LOAD	equ	.198	; load value for transmission
RTCC_TX_END	equ	.0	; load value at end of transmission (max)

RTCC_RX_LOAD	equ	.200	; load value for reception
RTCC_RX_START	equ	.194	; start value for reception (theoretical 190)
RTCC_RX_END	equ	.0	; load value at end of reception (max)

;-----------------------------------------------------------------------------
;miscellaneous values

portb_init_val	equ	0ffh	; initialiser for PORTB init/boot/trip value
csum_zero_rplc	equ	05ah	; zero checksum replacer

EE_CHECK	equ	(TS_SBY + TM_BOOT + TM_PSU)

;-----------------------------------------------------------------------------
;flash clock programming values

flash_unity	equ	011h	; unity value for flash programming

;*****************************************************************************
;EEPROM offsets - MUST match initialisation data below

ee_sby_tmr	equ	01h	; store for standby timer
ee_boot_tmr	equ	02h	; store for boot timer
ee_psu_tmr	equ	03h	; store for PSU timer
ee_checksum	equ	04h	; store for EEPROM checksum

ee_null_addr	equ	3fh	; null EEPROM address (write protect)

;-----------------------------------------------------------------------------
;EEPROM data storage - MUST match offsets above

	ORG     H'2100'
	de	0
	de	TS_SBY		;standby timer
	de	TM_BOOT		;boot timer
	de	TM_PSU		;psu override timer
	de	EE_CHECK	;EEPROM checksum value

;*****************************************************************************
;******************************  RESET VECTOR  *******************************
;*****************************************************************************

;-----------------------------------------------------------------------------
;reset vector - jump to main statup routine
	ORG     00
	goto	INIT

;*****************************************************************************
;****************************  INTERRUPT ROUTINE  ****************************
;*****************************************************************************

;*****************************************************************************
;
; IMPORTANT! In this system, INT0 is connected to the RS232 receive line from
; the PC, and is set to trigger whenever a start bit (ignored for data) is 
; detected. The timer interrupt is used for both transmission and reception.
; For reliable operation, the INT0 processor must preceed the timer overflow
; processor.
;
; Also note that the system will be more reliable if INT0 is NOT disabled
; during reception, after the first start bit has been detected. Though this
; might appear to be the best method, it actually worsens jitter, as the
; interrupt flag is set even when the interrupt is disabled, so that it will
; be processed (wasting 4 instruction cycles) when the next timer interrupt
; fires off. Keeping the interrupt enabled will mean that level changes will
; cause a seperate interrupt BETWEEN timer ticks, therefore not affecting
; the timing operation at all.
;
; Be careful to ensure that the interrupt routine instruction time does not
; exceed the availabe RTCC delay. The worst case is for transmission and
; reception, where only about 50 - 60 cycles are left after the timer has been
; reloaded. For this reason, keep all interrupt code as short as possible.
;
; Also note that if a start bit is detected while a timer interrupt is
; being serviced, the start bit detection will be delayed. This could
; potentially lead to comms errors, so it is vital to keep this routine
; as short as possible.
;
;*****************************************************************************

;interrupt routine - ensure that it does not clash with main code
	ORG     04

INTR	
;start of routine - preserve W reg and status
	movwf	W_TEMP		;save W in temp register
	swapf	STATUS, W	;store status register swapped in W
	bcf	STATUS, RP0	;change to bank 0
	movwf	S_TEMP		;save swapped status in temp register

;work out what interrupted us (order is VERY important)
	btfsc	INTCON, INTF	;test for external interrupt
	goto	I_INTF		;jump if bit set
E_INTF	
	btfsc	INTCON, T0IF	;test for timer overflow
	goto	I_T0IF		;jump if bit set
E_TOIF	
;restore W reg and status, and exit
	swapf	S_TEMP, W	;swap temp status back into W (original state)
	movwf	STATUS		;move W into status register
	swapf	W_TEMP, F	;swap contents of W_TEMP
	swapf	W_TEMP, W	;swap back into W (original state)
	retfie

;.............................................................................
;process INT signal int
I_INTF
;run start bit detect routine, if receive NOT currently progressing
	btfss	_rcvProgress	;check if RX progress SET
	goto	SBitDetected	;JUMP to start bit detect if CLEAR
INTFR
;exit INT signal int
	bcf	INTCON, INTF	;clear external interrupt flag
	goto	E_INTF		;jump back to correct place

;.............................................................................
;process Timer overflow int
I_T0IF
;now set timer overflow flag
	bsf	_tim_overflow	;set timer overflow flag

;run transmission routine, if transmission in progress (returns to ITOFR)
	btfsc	_txmtProgress	;check if progress CLEAR
	goto	TxNextBit	;JUMP to transmitter if SET

;run reception routine, if reception in progress (returns to ITOFR)
	btfsc	_rcvProgress	;check if progress CLEAR
	goto	RcvNextBit	;JUMP to receiver if SET

ITOF_IDLE
;timer idling, so decrement timer 1 and check if zero reached
        decfsz  CT_INT1,F	;decrement counter, checking if zero hit
        goto    ITOFR		;jump to end of routine (skip if zero)

;zero reached, so firstly reset timer 1
        movlw	TM_INTR1	;zero reached, so reload counter value
	movwf	CT_INT1		;store in interrupt counter
	
;now decrement timer 2 and check if zero reached
        decfsz  CT_INT2,F	;decrement counter, checking if zero hit
        goto    ITOFR		;jump to end of routine (skip if zero)

;zero reached, so firstly reset timer 2
        movlw	TM_INTR2	;zero reached, so reload counter value
	movwf	CT_INT2		;store in interrupt counter
	
;now decrement standby counter, unless result would become zero
	decfsz	CT_SBY, W	;decrement, but store in W, and check if zero
	movwf	CT_SBY		;store W back in counter (unless zero)

ITOFR
;exit timer int
	bcf	INTCON, T0IF	;clear timer overflow flag
	goto	E_TOIF		;jump back to correct place

;.............................................................................

;*****************************************************************************
;*****************************  MAIN CODE BODY  ******************************
;*****************************************************************************

;*****************************************************************************
;
; The system can be in one of four states:
;	1) Standby mode - entered via SBYSET
;	2) Boot Hold - entered via BHOLD_SET
;	3) Run mode - entered via FINAL_BOOT
;	4) Fail mode - entered via SYSFAIL
;
; In all modes, the system fully supports RS232 comms, and can report back
; its status at any time. It always initially boots into standby mode, but
; if the standby enable switch is disabled, it will quickly exit to run mode.
; To start from standby, a special "wake up" command should be sent to the
; board, as by default DTR is ignored (in order to prevent false starts).
; To force DTR compatibility, config switch 1 should be set open.
;
; The system will go into standby mode if the standby timer expires, assuming
; that the standby enable switch is enabled. As long as the switch is enabled,
; it can also be put into standby by a specific command; if the switch is 
; disabled, this command will simply be ignored.
;
; Note that if the standby timer is set to the maximum value (0xff), it is
; effectively disabled. The system can still be put into standby via specific
; commands, however.
;
; When the system is in failure mode, it will still respond to "sleep" and
; "wake up" commands, simply controlling the display reset as appropriate.
; This is useful for automatic fault diagnosis, if the panels are for some
; reason blocking QUICC-IDM's comms. When the panels are in reset, their
; comms are automatically disabled.
;
; When in failure mode, the only means of exiting is if a recover request is
; received, which effectively jumps back to standby mode.
;
; The "boot hold" mode is a special mode which keeps the PSU off and the
; panels out of reset. This mode is primarily used for diagnostic purposes, 
; though it can be used to generate a guarenteed clean boot. Note that it
; is transitionally passed through whenever a normal startup command 
; is received from standby. Also note that if it is entered from the main
; run mode, the "boot_portB" status is NOT updated, since at this point
; the panels are out of reset, and would therefore indicate a good watchdog
; status because the PSU is still outputting voltage.
;
; NOTE: If the unit resets due to a PIC watchdog trip, this routine waits
; waits for a LONG delay, to ensure the failure refgisters with the PC.
; However, before doing this, it puts the panels in reset and loads the
; initial portB status (obtains config switches). The reason for doing this
; is so that the PC can communicate with the unit during this delay, and
; the unit will be correctly configured. After this, the normal system
; initial store routine is run.
;
; The display reset request will briefly pulse the display reset line active
; for a short interval in ALL modes, EXCEPT standby mode where the displays
; are already in reset. In this mode, the "boot" LED will be pulsed instead,
; which could be used as an indication that the controlling PC is operational.
;
; The fault "lamp test" command will briefly illuminate the "fault" LED, 
; allowing it to be tested, as it will only otherwise illuminate if a genuine 
; fault is present. This will not have any effect if the system is already in
; fault reporting mode, for obvious reasons.
;
;*****************************************************************************

;-----------------------------------------------------------------------------
;start of main body of code - ensure that it does not clash with int routine
	ORG	40H

INIT
;ensure we are looking at first bank
	bcf	STATUS, RP0	;select first bank

;load initial values into ports, to prevent glitches
	movlw	PA_DFLT		;preset port A status
        movwf   PORTA
	movlw   PB_DFLT		;preset port B status
        movwf   PORTB

;clear down all status stores - very important
	clrf	SerialStatus
	clrf	M_FLAGS
	clrf	S_FLAGS

;work out if reset was due to WDT, and store status
	btfss	STATUS, NOT_TO	;check if /Timeout bit SET
	bsf	_pic_wdog_trip	;set watchdog flag if CLEAR

;if EEPROM write error detection worked, here would be point to check it!!!

;initialise port directions and option register
	clrf	TMR0		;clear down TMR0
	bsf	STATUS, RP0	;select second bank
	bcf	EECON1, WREN	;disable EEPROM writes
	movlw	PA_TRIS		;preset port A direction
        movwf   TRISA
	movlw   PB_TRIS		;preset port B direction
        movwf   TRISB
	clrwdt			;clear down WDT
        movlw   OPT_VAL		;preset option register 
        movwf	OPTION_REG
	bcf	STATUS, RP0	;select first bank

;now extract all global values from EEPROM (and set null EEPROM address)
	call	get_global_eeprom_data

;load the receive buffer head and tail pointers
	movlw	RxBuff		;load OFFSET of receive buffer
	movwf	RxHead		;copy this into head pointer
	movwf	RxTail		;copy this into tail pointer

;load high level values
	movlw	cm_wait_zero	;load initial comms mode
	movwf	comms_mode	;

;initialise miscellaneous stores
	movlw	portb_init_val	;load initial value 
	movwf	trip_portb	;into trip store
	movwf	boot_portb	;into boot store
	movwf	init_portb	;into init store

;load flash setters and counters
	movlw	flash_unity
	movwf	flash1_set
	movwf	flash2_set
	movwf	flash1_cnt
	movwf	flash2_cnt

;now preload all interrupt counters (use EEPROM loaded standby timer)
        movlw	TM_INTR1	;set fast interrupt timer value
	movwf	CT_INT1
        movlw	TM_INTR2	;set slow interrupt timer value
	movwf	CT_INT2
        movfw	sby_tmr		;set slow standby timer value
	movwf	CT_SBY

;unmask timer overflow and ext int, and enable global interrupts
	bcf	INTCON, RBIE	;mask RB port change int
	bcf	INTCON, EEIE	;mask EEPROM write complete int
	bsf	INTCON, INTE	;unmask external INT0 int
	bsf	INTCON, T0IE	;unmask timer overflow int
	bsf	INTCON, GIE	;FINALLY, enable global ints

;.............................................................................
;check if EEPROM checksum was bad (from global data load above)
	btfss	_bad_ee_check	;check if bad EEPROM checksum flag SET
	goto	INIT_CONT1	;skip next section if CLEAR

;bad EEPROM checksum, so illuminate fault LED, wait for delay, and extinguish
	bcf	PORTA, FAIL_LD	;illuminate the "fault" LED
	movlw	TM_IFAULT	;wait for init fault delay
	call	wait_delay
	bsf	PORTA, FAIL_LD	;extinguish the "fault" LED

;.............................................................................
INIT_CONT1
;if PIC reset due to a watchdog trip, wait for a longish delay
	btfss	_pic_wdog_trip	;check if PIC watchdog trip flag SET
	goto	INIT_CONT2	;skip next section if CLEAR

;PIC watchdog tripped, so illuminate boot LED and set flag
	bcf	PORTB, BOOT_LD	;illuminate the "boot" LED
	bsf	_pic_wdog_boot	;indicate watchdog re-boot in progress

;now, ensure panels in reset, wait for delay, and read INIT status
	bcf	PORTA, DP_RST	;ensure panels in reset
	movlw	TM_INITB	;wait for holdoff delay
	call	wait_delay
	movfw	PORTB		;take copy of port B
	movwf	init_portb	;store in initial store

;now wait for full holdoff delay
	movlw	TM_WHOLD	;wait for special holdoff delay
	call	wait_delay

;extinguish boot LED and clear flag, indicating ready for use
	bsf	PORTB, BOOT_LD	;extinguish the "boot" LED
	bcf	_pic_wdog_boot	;clear watchdog re-boot flag

;.............................................................................
INIT_CONT2
;firstly, take panels out of reset, wait for delay, and read BOOT status
	bsf	PORTA, DP_RST	;take the panels out of reset
	movlw	TM_BOOTB	;wait for holdoff delay
	call	wait_delay
	movfw	PORTB		;take copy of port B
	movwf	boot_portb	;store in boot store

;now, put panels back into reset, wait for delay, and read INIT status
	bcf	PORTA, DP_RST	;put the panels back in reset
	movlw	TM_INITB	;wait for holdoff delay
	call	wait_delay
	movfw	PORTB		;take copy of port B
	movwf	init_portb	;store in initial store

;get the correct stored DTR state, for initialisation
	call	check_dtr

;.............................................................................
;this is the entry point for going into standby (if not already)

SBYSET
;we should now setup standby mode (PSU, Boot & Fail off, displays reset)
	bcf	PORTA, PSU_ON	;switch off the PSU
	bsf	_sby_status	;indicate in standby
	bcf	PORTA, DP_RST	;put the display panels in reset
	bsf	PORTB, BOOT_LD	;extinguish the "boot" LED
	bsf	PORTA, FAIL_LD	;extinguish the "fault" LED

;wait for a short delay to prevent standby switch bounce
	movlw	TM_START		;wait for holdoff delay
	call	wait_delay

;.............................................................................
;STATE: we are now running in standby mode
;note that to achieve full boot, start request flag MUST be set

SBYCHK
;clear down appropriate status flags
	bcf	_sby_request	;clear down standby request

;firstly, check if the standby switch is enabled
	btfsc	PORTB, SBY_SW	;check if standby switch CLOSED
	bsf	_start_request	;trigger FULL startup if OPEN - IMPORTANT!

;now process any incoming comms
	call	handle_comms	;handle incoming comms
	clrwdt			;clear down WDT

;check if we've received a specific startup request or boot hold request
	btfsc	_start_request	;check if start request CLEAR
	goto	BHOLD_SET	;jump to standby exit if SET
	btfsc	_boot_hold_req	;check if boot hold request CLEAR
	goto	BHOLD_SET	;jump to standby exit if SET

;now check if in "DTR compatibility mode", config switch 1 OPEN
	btfss	_cfg_sw1 	;check if config switch 1 is OPEN
	goto	SBCONT1		;continue, ignoring DTR if CLOSED

;config switch 1 open, so check if change in DTR
	call	check_dtr	;check DTR, returns TRUE or FALSE
	iorlw	0		;get Z flag from result
	bz	SBCONT1		;jump to continue if NO CHANGE

;DTR changed, so set start request and jump to standby exit
	bsf	_start_request	;trigger FULL startup - IMPORTANT!
	goto	BHOLD_SET	;jump to standby exit

SBCONT1
;check if display reset request received
	btfss	_dreset_req	;check if display reset request SET
	goto	SBCONT2		;skip next bit if CLEAR

;as we're in standby mode, pulse the "boot" LED instead of reset line
	bcf	PORTB, BOOT_LD	;illuminate "boot" LED
	movlw	TM_DRESET	;wait for display reset delay
	call	wait_delay
	bsf	PORTB, BOOT_LD	;extinguish "boot" LED
	bcf	_dreset_req	;clear reset request

SBCONT2
;check if fault lamp test request received
	btfss	_flamp_test_req	;check if fault lamp test request SET
	goto	SBCONT3		;skip next bit if CLEAR

;simply pulse the "fault" LED for a short interval
	bcf	PORTA, FAIL_LD	;illuminate the "fault" LED
	movlw	TM_IFAULT	;wait for fault LED illuminate delay
	call	wait_delay
	bsf	PORTA, FAIL_LD	;extinguish the "fault" LED
	bcf	_flamp_test_req	;clear fault lamp test request

SBCONT3
;perform misc actions, and loop round to beginning
	bcf	PORTA, PSU_ON	;ENSURE PSU kept off
	bcf	PORTA, DP_RST	;ENSURE display panels in reset
	goto	SBYCHK		;loop round

;.............................................................................
;this is the boot hold entry point from the standby mode
;note the start request flag MUST NOT be cleared at this point

BHOLD_SET
;switch off the PSU, change flags and indicate panels booting
	bcf	PORTA, PSU_ON	;switch off the PSU
	bsf	_panel_boot	;indicate panels booting
	bcf	PORTB, BOOT_LD	;illuminate "boot" LED
	bcf	_sby_status	;indicate out of standby
	bsf	_sys_started	;indicate started from reset

;now take panels out of reset, wait for special delay, and store status
	bsf	PORTA, DP_RST	;take the panels out of reset
	movlw	TM_BOOTB	;wait for holdoff delay
	call	wait_delay
	movfw	PORTB		;take copy of port B
	movwf	boot_portb	;store in boot store

;.............................................................................
;this is the boot hold entry point from the main run mode, and also
;forms part of the boot hold setup from other modes

BHOLD_MAIN_SET
;switch off PSU, illuminate boot LED, and change flags
	bcf	PORTA, PSU_ON	;switch off the PSU
	bsf	_panel_boot	;indicate panels booting
	bcf	PORTB, BOOT_LD	;illuminate "boot" LED

;now wait for full boot delay
	movfw	boot_tmr	;wait for panel boot delay
	call	wait_delay

;.............................................................................
;STATE: we are now running in BOOT HOLD mode
;note that if in DTR compatibility mode, DTR must be checked

BHOLD_LOOP
;clear down appropriate status flags
	bcf	_boot_hold_req	;clear down boot hold request

;process any incoming comms
	call	handle_comms	;handle incoming comms
	clrwdt			;clear down WDT

;check if system start request received (either now or when standby exited)
	btfsc	_start_request	;check if start request CLEAR
	goto	FINAL_BOOT	;jump to final boot if SET

;check if display reset request received
	btfss	_dreset_req	;check if display reset request SET
	goto	BHCONT1		;skip next bit if CLEAR

;reset the display panels for fixed interval
	bcf	PORTA, DP_RST	;put display panels in reset
	movlw	TM_DRESET	;wait for display reset delay
	call	wait_delay
	bsf	PORTA, DP_RST	;take display panels out of reset
	movlw	TM_RRECOV	;now wait for panels to recover
	call	wait_delay
	bcf	_dreset_req	;clear reset request

BHCONT1
;now check if standby switch enabled
	btfsc	PORTB, SBY_SW	;check if standby switch CLOSED
	goto	BHCONT2		;switch OPEN, so skip ALL standby checks

;check if we've received a specific standby request
	btfsc	_sby_request	;check if standby request CLEAR
	goto	SBYSET		;jump to standby if SET

BHCONT2
;now check if in "DTR compatibility mode", config switch 1 OPEN
	btfss	_cfg_sw1 	;check if config switch 1 is OPEN
	goto	BHCONT3		;continue, ignoring DTR if CLOSED

;config switch 1 open, so check if change in DTR
	call	check_dtr	;check DTR, returns TRUE or FALSE
	iorlw	0		;get Z flag from result
	bnz	FINAL_BOOT	;jump to final boot if CHANGE

BHCONT3
;check if fault lamp test request received
	btfss	_flamp_test_req	;check if fault lamp test request SET
	goto	BHCONT4		;skip next bit if CLEAR

;simply pulse the "fault" LED for a short interval
	bcf	PORTA, FAIL_LD	;illuminate the "fault" LED
	movlw	TM_IFAULT	;wait for fault LED illuminate delay
	call	wait_delay
	bsf	PORTA, FAIL_LD	;extinguish the "fault" LED
	bcf	_flamp_test_req	;clear fault lamp test request

BHCONT4
;perform misc operations and loop round to beginning
	bcf	_sby_request	;clear down standby request
	goto	BHOLD_LOOP	;loop round

;.............................................................................
;system now fully starting - switch on PSU and change flags

FINAL_BOOT
;now turn on PSU, and wait for PSU override delay
	bsf	PORTA, PSU_ON	;switch on the PSU
	movfw	psu_tmr		;wait for PSU on delay
	call	wait_delay

;finally, extinguish the "boot" LED, reload the standby timer, and exit
	bsf	PORTB, BOOT_LD	;extinguish "boot" LED
	bcf	_panel_boot	;indicate panels finished booting
	movfw	sby_tmr		;reset slow standby timer to max
	movwf	CT_SBY		;

;.............................................................................
;STATE: main loop - panels are now active

MAIN
;clear down appropriate status flags
	bcf	_start_request	;clear down start request

;check the display watchdog status
	call	check_wdog	;check watchdog, returning TRUE if OK
	iorlw	0		;get Z flag from result
	bz	SYSFAIL		;jump to failure routine if fault

;process any incoming comms
	call	handle_comms	;handle incoming comms
	clrwdt			;clear down WDT

;run DTR checking routine, which also controls flash clocks
	call	check_dtr	;check DTR, resetting standby timer if changed

;check if display reset request received
	btfss	_dreset_req	;check if display reset request SET
	goto	MNCONT1		;skip next bit if CLEAR

;reset the display panels for fixed interval
	bcf	PORTA, DP_RST	;put display panels in reset
	movlw	TM_DRESET	;wait for display reset delay
	call	wait_delay
	bsf	PORTA, DP_RST	;take display panels out of reset
	movlw	TM_RRECOV	;now wait for panels to recover
	call	wait_delay
	bcf	_dreset_req

MNCONT1
;check if we've received a specific boot hold request
	btfsc	_boot_hold_req	;check if boot hold request CLEAR
	goto	BHOLD_MAIN_SET	;jump to boot hold main entry point if SET

;now check if standby switch enabled
	btfsc	PORTB, SBY_SW	;check if standby switch CLOSED
	goto	MNCONT3		;switch OPEN, so skip ALL standby checks

;check if we've received a specific standby request
	btfsc	_sby_request	;check if standby request CLEAR
	goto	SBYSET		;jump to standby if SET

;check if standby timer disabled (maximum timer setting 0xff)
	incfsz	sby_tmr, W	;increment SETTING, store in W, check if zero
	goto	MNCONT2		;continue if non zero
	goto	MNCONT3		;zero, so skip standby timer check

MNCONT2
;now check if standby timer expired
	decfsz	CT_SBY, W	;decrement, but store in W, and check if zero
	goto	MNCONT3		;continue if non zero
	goto	SBYSET		;timer expierd, so jump to standby setter

MNCONT3
;check if fault lamp test request received
	btfss	_flamp_test_req	;check if fault lamp test request SET
	goto	MNCONT4		;skip next bit if CLEAR

;simply pulse the "fault" LED for a short interval
	bcf	PORTA, FAIL_LD	;illuminate the "fault" LED
	movlw	TM_IFAULT	;wait for fault LED illuminate delay
	call	wait_delay
	bsf	PORTA, FAIL_LD	;extinguish the "fault" LED
	bcf	_flamp_test_req	;clear fault lamp test request

MNCONT4
;perform misc operations and loop round to beginning
	bcf	_sby_request	;clear down standby request
	goto	MAIN		;loop round

;.............................................................................
;fault handler - shutdown PSU and wait for healthy watchdogs

SYSFAIL
;immediately illuminate the fault LED (aids diagnostics)
	bcf	PORTA, FAIL_LD	;illuminate the "fault" LED

;wait for a short delay, to verify that system has really failed
	movlw	TM_FAIL1	;wait for failure delay
	call	wait_delay

;check the watchdog status, and jump back if recovered
	bsf	PORTA, FAIL_LD	;extinguish the "fault" LED
	call	check_wdog	;check watchdog, returning TRUE if OK
	iorlw	0		;get Z flag from result
	bnz	MAIN		;jump back to main loop if recovered

;really faulty, so take a copy of port B status, in trip store
	movfw	PORTB		;take copy of port B
	movwf	trip_portb	;store in trip store

;now shut down PSU, clear any recover request, and indicate fault
	bcf	PORTA, PSU_ON	;switch the PSU off
	bcf	_recover_req	;clear the recover request flag
	bcf	PORTA, FAIL_LD	;illuminate the "fault" LED
	bsf	_fault_status	;set the fault status flag

;.............................................................................
;STATE: we are running in fault reporting mode

SFLASH
;failed, so flash the failure LED, checking comms in the meantime
	movlw	TM_FLASH	;wait for LED flash delay
	call	wait_delay
	bsf	PORTA, FAIL_LD	;extinguish the "fault" LED
	movlw	TM_FLASH	;wait for LED flash delay
	call	wait_delay
	bcf	PORTA, FAIL_LD	;illuminate the "fault" LED

;perform misc operations for stability
	bcf	_boot_hold_req	;clear boot hold request
	bcf	PORTA, PSU_ON	;ENSURE PSU kept off
	bcf	_flamp_test_req	;clear fault lamp test request

;check if standby request received
	btfsc	_sby_request	;check if standby request CLEAR
	bcf	PORTA, DP_RST	;put panels in reset if SET
	btfsc	_sby_request	;check if standby request CLEAR
	bsf	_sby_status	;indicate in standby if SET
	bcf	_sby_request	;clear standby request

;check if start request received
	btfsc	_start_request	;check if start request CLEAR
	bsf	PORTA, DP_RST	;take panels out of reset if SET
	btfsc	_start_request	;check if start request CLEAR
	bcf	_sby_status	;indicate out of standby if SET
	bcf	_start_request	;clear start request

;check if recover request received
	btfsc	_recover_req	;check if recover request CLEAR
	bcf	_fault_status	;clear the fault status flag if SET
	btfsc	_recover_req	;check if recover request CLEAR
	goto	SBYSET		;jump out to standby if SET

;now check if display reset request received
	btfsc	_dreset_req	;check if display reset request CLEAR
	bcf	PORTA, DP_RST	;put display panels in reset if SET
	btfsc	_dreset_req	;check if display reset request CLEAR
	bsf	_start_request	;trigger start request on next cycle if SET
	bcf	_dreset_req	;clear reset request

;loop round to beginning
	goto	SFLASH		;loop round forever


;*****************************************************************************
;************************  LOW LEVEL SUBROUTINES  ****************************
;*****************************************************************************

;*****************************************************************************
;				wait_delay
;
; Waits for specified delay in multiples of 20ms, or thereabouts.
; Maximum delay is approximately 5 seconds.
; Calls check_dtr during operation, to maintain flash clocks.
; Note that the timer runs at a different rate when the serial comms is
; active, so this routime attempts to compensate. However, there will still
; be variations in the delay time, so this cannot be used for really
; accurate time intervals.
;
; The routine uses 2 counters:
;	COUNT1 - converts raw timer interrupt to 20ms intervals
;	COUNT2 - converts COUNT1 to user-specified amount
;
; Note that this routine runs the DTR checker and the comms handler from
; within the inner loop. The watchdog timer is reset from within the OUTER
; loop, as this will cause a reset if timer interrupts fail.
;
;*****************************************************************************

wait_delay
;take in W register and wait for delay
	movwf   COUNT2          ;store W in COUNT2 temporarily

WDEL1
;counter 1 loop - reloads counter 1 to wait for another 20ms interval
	movlw	TL_CNT1		;load counter 1 with start val
	movwf   COUNT1		;

;clear down watchdog in this OUTER loop, to generate reset if interrupts fail
	clrwdt			;clear down WDT - IMPORTANT!

WDEL0
;interrupt timer wait loop - call DTR checker and comms handler
	call	check_dtr	;call DTR checker, to maintain flash clocks
	call	handle_comms	;call comms handler

;check if timer overflow bit set
	btfss	_tim_overflow	;check if timer overflow bit set
        goto    WDEL0		;bit clear, so loop round

;interrupt timer triggered, so clear bit, and obtain correct decrement
	bcf     _tim_overflow	;clear overflow flags bit
	movlw	TDEC_IDLE	;PRE-LOAD delay counter with idle decrement
	btfsc	_rcvProgress	;check if RX pregress CLEAR
	movlw	TDEC_COMMS	;load comms value if SET
	btfsc	_txmtProgress	;check if TX pregress CLEAR
	movlw	TDEC_COMMS	;load comms value if SET

;SUBTRACT decrement , checking if gone negative (ie. expired)
	subwf	COUNT1,F	;subtract decrement from COUNT1
	bc	WDEL0		;loop round to int checkier if POSITIVE

;expired, so decrement counter 2, and loop round if not zero
        decfsz  COUNT2,F        ;decrement counter, checking for zero
        goto    WDEL1		;not zero, so loop round

;delay expired, so exit
	return

;*****************************************************************************
;				handle_flash
;
; Handles display flash clocks, simply following DTR if settings are both
; unity, or counting if either setting is greater. Flash clocks 1 and 2 are
; handled completely independantly.
;
; Note that for code efficiency, this routine will write out a preset value,
; perform a check, and invert it if required. Though this could theoretically
; lead to spikes, it will not in practice as the values are only written out
; when being CHANGED from one state to another.
;
; Also note that with default hardware settings, flash 1 and flash 2 are
; inverted from one another, so the SAME values are written to get antiphase
; effects.
;
;*****************************************************************************

handle_flash
;for flash clock 1, check if counters are unity
	movlw	flash_unity	;get flash unity value
	subwf	flash1_set, W	;subtract from setting, storing in W
	bnz	hfl1_count	;jump to next section if non-unity
	
;unity clocks, so simply follow DTR status
	bsf	PORTA, FLASH1	;PRE-SET flash clock 1
	btfss	_dtr_status	;check if DTR status SET
	bcf	PORTA, FLASH1	;CLEAR flash clock 1 if CLEAR
	goto	hfl1_end	;jump to end of flash 1

hfl1_count
;need to handle flash clock counters, so select which one
	btfss	PORTA, FLASH1	;check if flash clock currently SET
	goto	hfl1_off	;jump to next section if CLEAR

;currently on, so decrement the on counter and check if zero
	movlw	010h		;decrement upper nybble
	subwf	flash1_cnt, F	;subtract from counter
	movfw	flash1_cnt	;get counter value in W
	andlw	0f0h		;mask out lower bits
	bnz	hfl1_end	;jump to end if non-zero

;counter expired, so reload and turn OFF
	movfw	flash1_set	;load set value in W
	movwf	flash1_cnt	;put this into counter
	bcf	PORTA, FLASH1	;CLEAR flash clock 1
	goto	hfl1_end	;jump to end of flash 1

hfl1_off
;currently off, so decrement the off counter
	movlw	001h		;decrement lower nybble
	subwf	flash1_cnt, F	;subtract from counter
	movfw	flash1_cnt	;get counter value in W
	andlw	00fh		;mask out upper bits
	bnz	hfl1_end	;jump to end if non-zero

;counter expired, so reload and turn OFF
	movfw	flash1_set	;load set value in W
	movwf	flash1_cnt	;put this into counter
	bsf	PORTA, FLASH1	;SET flash clock 1

hfl1_end
;............................................................................
;for flash clock 1, check if counters are unity
	movlw	flash_unity	;get flash unity value
	subwf	flash2_set, W	;subtract from setting, storing in W
	bnz	hfl2_count	;jump to next section if non-unity
	
;unity clocks, so simply follow DTR status
	bsf	PORTA, FLASH2	;PRE-SET flash clock 2
	btfss	_dtr_status	;check if DTR status SET
	bcf	PORTA, FLASH2	;CLEAR flash clock 2 if CLEAR
	goto	hfl2_end	;jump to end of flash 2

hfl2_count
;need to handle flash clock counters, so select which one
	btfss	PORTA, FLASH2	;check if flash clock currently SET
	goto	hfl2_off	;jump to next section if CLEAR

;currently on, so decrement the on counter and check if zero
	movlw	010h		;decrement upper nybble
	subwf	flash2_cnt, F	;subtract from counter
	movfw	flash2_cnt	;get counter value in W
	andlw	0f0h		;mask out lower bits
	bnz	hfl2_end	;jump to end if non-zero

;counter expired, so reload and turn OFF
	movfw	flash2_set	;load set value in W
	movwf	flash2_cnt	;put this into counter
	bcf	PORTA, FLASH2	;CLEAR flash clock 2
	goto	hfl2_end	;jump to end of flash 2

hfl2_off
;currently off, so decrement the off counter
	movlw	001h		;decrement lower nybble
	subwf	flash2_cnt, F	;subtract from counter
	movfw	flash2_cnt	;get counter value in W
	andlw	00fh		;mask out upper bits
	bnz	hfl2_end	;jump to end if non-zero

;counter expired, so reload and turn OFF
	movfw	flash2_set	;load set value in W
	movwf	flash2_cnt	;put this into counter
	bsf	PORTA, FLASH2	;SET flash clock 2

hfl2_end
;............................................................................
;exit the routine
	return

;*****************************************************************************
;				check_dtr
;
; Check for DTR change, storing new state. If changed, resets standby timer.
; Returns TRUE if DTR has changed this cycle, otherwise FALSE.
;
;*****************************************************************************

check_dtr
;firstly, check DTR status, and act accordingly
	btfss	PORTB, PC_DTR 	;check if DTR is SET
        goto    CDCLR		;jump to clear routine if CLEAR

;.............................................................................
;DTR is set, so check if store is CLEAR, and update if so
	btfsc	_dtr_status	;check if store is CLEAR
	retlw	FALSE		;store SET, so directly exit

;DTR has CHANGED to SET, so update store and flash clocks
	bsf	_dtr_status	;SET the flag in the store
	call	handle_flash	;handle flash clocks
	goto	CDCHNG		;exit the routine, with TRUE

;.............................................................................
;DTR is clear, so check if store is SET, and update if so
CDCLR	
	btfss	_dtr_status	;check if store is SET
	retlw	FALSE		;store CLEAR, so directly exit

;DTR has CHANGED to CLEAR, so update store and flash clocks
	bcf	_dtr_status	;CLEAR the flag in the store
	call	handle_flash	;handle flash clocks

;.............................................................................
;reset standby timer, then exit indicating change of state
CDCHNG
	movfw	sby_tmr		;reset slow standby timer to max
	movwf	CT_SBY
	retlw	TRUE		;indicate DTR state change

;*****************************************************************************
;				check_wdog
;
; Check watchdog status. Returns TRUE if all healthy (high), otherwise FALSE.
;
;*****************************************************************************

check_wdog
;check each watchdog in turn
	btfss	PORTB, WDOG1 	;check if watchdog 1 is SET
	retlw	FALSE		;watchdog CLEAR, so directly exit
	btfss	PORTB, WDOG2 	;check if watchdog 2 is SET
	retlw	FALSE		;watchdog CLEAR, so directly exit
	btfss	PORTB, WDOG3 	;check if watchdog 3 is SET
	retlw	FALSE		;watchdog CLEAR, so directly exit
	retlw	TRUE		;indicate all healthy

;*****************************************************************************
;************************  RS232 TRANSMIT SUBROUTINES  ***********************
;*****************************************************************************

;*****************************************************************************
;				PutChar
;
;	Function to transmit A Byte Of Data, supplied in W register.
;
; IMPORTANT! Transmit/Receive arbitration: 
;
; This routine firstly sets the transmission request flag, which will prevent
; the receiver from triggering on a start bit. It then waits until both the
; transmitter and receiver are idle before initiating the transmission proper.
; Once the transmission progress flag is set, the request flag is cleared,
; as the receiver will also ignore triggers when a transmit is in progress.
; Note that the sequence of events is CRITICAL, as both timer and receive 
; trigger interrupts can occur at any time. This order is:
;	
; 1) Set transmit request flag
; 2) WAIT until both receiver and transmitter are idle
; 3) Set up parameters for transmission
; 4) Activate transmitter progress flag
; 5) Clear the transmit request flag
;
; As you can see from the above, transmission takes PRIORITY over reception.
; Received data will be thrown away once a transmit is initiated. This is
; essential because the receiver has more effect on the RTCC than the 
; transmitter, which simply picks up on the next RTCC interrupt. To avoid
; data being lost, the PC should not transmit ANY data between issuing
; a card command/status request and getting a response.
;
; NOTE: This routine can be called from within the delay routine, so must
; never call this delay routine!
;
;*****************************************************************************

PutChar:
; firstly, SET the transmission request flag
	bsf	_txmtRequest			;SET the transmission request

_WaitPut
; now, check that reception or transmission NOT already in progress
	btfsc	_rcvProgress			;check if RX pregress CLEAR
	goto	_WaitPut			;loop round if SET
	btfsc	_txmtProgress			;check if TX pregress CLEAR
	goto	_WaitPut			;loop round if SET

;now pre-load all transmission parameters
	movwf	HoldReg				; store specified character
	movlw	TX_BITC_LOAD			; pre-load bitcount
	movwf	BitCount
	bsf	_txmtStart			; transmission start enable

;FINALLY, trigger transmission on next timer interrupt
	bsf	_txmtProgress			; transmission in progress
	bcf	_txmtRequest			;CLEAR transmission request
	retlw	TRUE				; indicate started

;*****************************************************************************
;		TxNextBit - Internal Subroutine
; entered from Interrupt Service Routine when transmission enabled
;
; NOTE: After the last stop bit has bee sent, the transmit in progress flag
; will be cleared, but another RTCC interval will have been set. This is
; necessary, because the stop bit must be set for at least one bit interval to
; register. This will not cause a problem, because even if another 
; transmission is immediately initiated, the last RTCC interval will have
; to expire, and if a character is received, the RTCC will be reset anyway.
; To increase reliability, a special "long" interval is set for the stop bit.
;
; Also note that for maximum code efficiency, this routine is JUMPED to, 
; and JUMPS directly back to the end of the interrupt routine. This looks
; less clean than calls, but results in smaller and faster code.
;
; IMPORTANT! In order to balance the timing, so that each bit is of the
; correct width, extra NOP instructions are inserted in the start and stop
; bit routines, so as to delay them by the same amount as the data routines.
;
;
;*****************************************************************************

TxNextBit:
;firstly, set up timer for NEXT trigger
	movlw	RTCC_TX_LOAD		; Prepare NEXT Interrupt
	movwf	TMR0

;----------------------------------------------------------------------------
;now, check if just started transmission
	btfsc	_txmtStart		;check if start CLEAR
	goto	_SetStartBit		;skip to next section if SET

;transmission in progress, so check bit counter
	decfsz	BitCount, F		;decrement bitcount register
	goto	_NextTxmtBit		;TX next bit if NOT zero
;
_StopBit
;counter expired, so transmit stop bit and finish
	nop				;delay to balance timing
	nop				;
	nop				;
	nop				;
	bsf	TX_PIN			; Stop Bit is HIGH
	movlw	RTCC_TX_END		; Prepare LONG interrupt
	movwf	TMR0			;
	bcf	_txmtProgress		;clear progress indicator
        goto    ITOFR			;jump to end of int
;
_NextTxmtBit
;transmit the next bit, and exit routine
	rrf	HoldReg, F		;shift lowest bit thru carry
	skpc				;skip next if SET
        bcf	TX_PIN			;output low if CLEAR
	skpnc				;skip next if CLEAR
	bsf	TX_PIN			;output high if SET
        goto    ITOFR			;jump to end of int

_SetStartBit
;request start bit, so set it and exit
	nop				;delay to balance timing
	nop				;
	nop				;
	nop				;
	nop				;
	bcf	TX_PIN			;Start bit is LOW
	bcf	_txmtStart		; clear start bit
        goto    ITOFR			;jump to end of int

;*****************************************************************************
;*************************  RS232 RECEIVE SUBROUTINES  ***********************
;*****************************************************************************

;*****************************************************************************
;				GetChar
;
; Gets next character from receive buffer, if available, otherwise returns 0.
; This function does not affect the _rcvOver bit.
; Before calling this routine, you should call the CharCount function to 
; determine how many characters are present. The output character is also
; present in global RxStore, as well as the W register (and Z flag is 
; affected)
;
;*****************************************************************************

GetChar:
;check if head and tail pointers match, and exit with zero if so
	movfw	RxHead		;get contents of head pointer
	subwf	RxTail, W	;subtract from tail, but place result in W
	bz	_GetCExit	;if zero, directly exit

;character available, so extract from tail pointer
	movfw	RxTail		;get contents of head pointer
	movwf	FSR		;store in FSR, for indirect addr
	movfw	INDF		;extract character from receive buffer
	movwf	RxStore		;store output char in RxStore

;increment tail pointer, and adjust if over limit
	incf	RxTail		;increment tail pointer
	movlw	RxBuffEnd	;get OFFSET of buffer end
	subwf	RxTail, W	;subtract, but place result in W
	bnc	_GetCFin	;skip next bit if result NEGATIVE

;hit limit, so reset head pointer
	movlw	RxBuff		;load OFFSET of receive buffer
	movwf	RxTail		;copy this into tail pointer

_GetCFin
;finishing normally, so get the output character in the W reg
	movfw	RxStore		;get output char in W

_GetCExit
	return			;directly return

;*****************************************************************************
;				CharCount
;
; Calculates how many characters are in the buffer, returning this in W
; Flags are also guarenteed to be correct, allowing fast compares
;
;*****************************************************************************

CharCount:
;check if head and tail pointers match, and exit with zero if so
	movfw	RxTail		;get contents of tail pointer
	subwf	RxHead, W	;subtract from head, but place result in W
	bz	_CharCExit	;if zero, directly exit

;now check carry flag (clear means Tail > Head, set means head > tail)
;if head > tail, then W contains number of chars in buffer, so we can exit
	bc	_CharCExit	;exit if carry SET (head > tail)

;tail > head, so we need to subtract from buffer length to get num chars
	movfw	RxHead		;get contents of head pointer
	subwf	RxTail, W	;subtract from tail, but place result in W
	sublw	rx_buff_size	;subtract result from buffer size

_CharCExit
	return			;directly return

;*****************************************************************************
;		SBitDetected - Internal Subroutine
; Entered from INT0 Interrupt Service Routine when Start Bit Is detected.
;
; IMPORTANT! Transmit/Receive arbitration: 
;
; This routine is the entry point for reception, but will only initiate a
; reception cycle if (in this order):
;
; 1) A reception is NOT already in progress (checked by caller)
; 2) A transmit request is NOT pending
; 3) A transmission is NOT already in progress
; 4) The receive pin logic level indicates a start bit
;
; As long as these conditions are satisfied, the routine assumes a valid
; start bit, and resets the RTCC to time halfway into the first start bit.
; This is the ONLY place where the RTCC is ASYNCHRONOUSLY changed; in other
; places, it may be tweaked, but this is within the timer interrupt itself.
; For this reason, received characters will affect any timing loops being
; performed by the system. This is unavoidable.
;
; To avoid the situation where an RTCC interrupt has already triggered, this
; routine will clear the RTCC overflow bit if it is initiating a reception.
; This prevents the first bit sample from occurring immediately! For this
; reason (and to prevent unpredicatable start delays), the INT0 check MUST
; preceed the timer overflow check. 
;
; Also note that for maximum code efficiency, this routine is JUMPED to, 
; and JUMPS directly back to the end of the interrupt routine. This looks
; less clean than calls, but results in smaller and faster code.
;
;*****************************************************************************

SBitDetected:
;firstly, check if transmit request or pending 
	btfsc	_txmtProgress	;check if TX progress CLEAR
	goto	INTFR		;return if SET
	btfsc	_txmtRequest	;check if TX request CLEAR
	goto	INTFR		;return if SET

;now check that a valid start bit really was received
	btfsc	RX_PIN		; Check that pin CLEAR (start bit)
	goto	INTFR		; return directly if SET (false start bit)

;recption allowed, so adjust RTCC for FIRST bit detect
	movlw	RTCC_RX_START	; Prepare NEXT Interrupt
	movwf	TMR0

;----------------------------------------------------------------------------
;load in the receive bit count
	movlw	RX_BITC_LOAD	; pre-load bitcount
	movwf	BitCount

;set receive progress, clear any timer flag, and exit
	bsf	_rcvProgress	;set progress indication
	bcf	INTCON, T0IF	;clear timer overflow flag
	goto	INTFR		;return directly

;*****************************************************************************
;			RcvNextBit - Internal Subroutine
; Entered from Timer Interrupt Service Routine when Next bit to be read.
;
; When the outine has received a full character, it stores it in the
; receive buffer, setting the flag _rcvOver.
; Note that no checking is done on whether the buffer has any room in it,
; so that characters will be overwritten if it is not emptied fast enough.
; This is no worse than checking the buffer and throwing away the input 
; character if there was no room!
;
; Also note that for maximum code efficiency, this routine is JUMPED to, 
; and JUMPS directly back to the end of the interrupt routine. This looks
; less clean than calls, but results in smaller and faster code.
;
; IMPORTANT! This routine uses FS_TEMP to preserve the original FSR value.
; This ensures that this interrupt routine does not corrupt the FSR being
; used by the normal main loop code. Note that this is a recent bugfix,
; though no problems have ever been experienced with QUICC-IDM in this 
; respect.
;
;*****************************************************************************

RcvNextBit:
;firstly, set up timer for NEXT trigger
	movlw	RTCC_RX_LOAD		; Prepare NEXT Interrupt
	movwf	TMR0

;read bit from port and store in temp flag
	bcf	_rcvTemp		;PRE-clear temp bit
	btfsc	RX_PIN			;check if data bit CLEAR
	bsf	_rcvTemp		;set temp bit if SET

;----------------------------------------------------------------------------
;decrement bit counter, and check if at end
	decfsz	BitCount, F		;decrement bitcount register
	goto	_NextRcvBit		;obtain next bit if NOT zero	

_RcvStopBit
;counter expired, so check for framing error
	bcf	_FrameErr		;PRE-CLEAR framing error
	btfss	_rcvTemp		;check if (temp stored) stop bit SET
	bsf	_FrameErr		;set framing error if CLEAR

;now transfer holding buffer into single output register, and receive buffer
	movfw	FSR			;store FSR temporarily
	movwf	FS_TEMP			;
	movfw	RxHead			;get contents of head pointer
	movwf	FSR			;store in FSR, for indirect addr
	movfw	HoldReg			;get assembled HOLDING reg into WREG
	movwf	INDF			;store in receive buffer indirectly
	movfw	FS_TEMP			;restore original FSR value
	movwf	FSR			;

;increment head pointer, and adjust if over limit
	incf	RxHead			;increment head pointer
	movlw	RxBuffEnd		;get OFFSET of buffer end
	subwf	RxHead, W		;subtract, but place result in W
	bnc	_RcvBitFin		;skip next bit if result NEGATIVE

;hit limit, so reset head pointer
	movlw	RxBuff			;load OFFSET of receive buffer
	movwf	RxHead			;copy this into head pointer

_RcvBitFin
;adjust flags, and exit routine
	movlw	RTCC_RX_END		;Prepare LONG interrupt
	movwf	TMR0			;
	bcf	_rcvProgress		;clear progress indication
	bsf	_rcvOver		;Indicate byte received
        goto    ITOFR			;jump to end of int
;
_NextRcvBit
;ready for next bit, so adjust carry and shift into holding reg
	clrc				;PRE-clear carry
	btfsc	_rcvTemp		;check if (temp stored) data bit CLEAR
	setc				;set carry if SET
	rrf	HoldReg, F		;shift into holding register	
        goto    ITOFR			;jump to end of int


;*****************************************************************************
;************************  HIGH LEVEL SUBROUTINES  ***************************
;*****************************************************************************

;*****************************************************************************
;				handle_comms
;
; Gets next byte from the comms stream, if available, and performs actions
; depending on its value, and the current comms mode. Available modes are:
;
;	0 - waiting for 0 to be received - other values discarded.
;	1 - got zero - waiting for 0xFE startup header.
;	2 - got 0xFE startup header - waiting for command code 1
;	3 - got command - waiting for data value
;	4 - got data val - waiting for checksum of (command + data)
;	5 - got checksum - waiting for 0xEF term code (actions command)
;	
; Note that if zero is received during any mode, mode 1 will be selected.
; If invalid data is received, mode 0 will be selected.
;
; Command codes 1 and 2 provide some protection against currupted data.
; If they match, then the command will be executed, otherwise it will
; be discarded.
;
; After executing a valid command, a status message will be sent back to the
; PC. If a corrupted message is received, nothing is sent back. The status
; message is the same regardless of the command received.
;
; NOTE: This routine has been implemented as compares and branches, rather 
; than as a call table, because the latter requires interrupts to be disabled.
; In this system, the RTCC interrupt and INT0 interrupt are fairly critical
; as far as timing goes, so they should not be disabled unless absolutely
; necessary.
;
; NOTE2: Command and data codes are added to form a checksum. If this would 
; become zero, it is replaced with a special value, to prevent zeros ever
; having to be sent.
;
; NOTE3: This routine can be called from within the delay routine, so must
; never call this delay routine!
;
; NOTE4: Because this routine is executed frequently, it reloads the TRIS
; and OPTION registers, in order to protect against a brown-out condition.
;
;*****************************************************************************

handle_comms:
;reload option and TRIS registers - IMPORTANT!
	bsf	STATUS, RP0	;select second bank
	movlw	PA_TRIS		;preset port A direction
        movwf   TRISA
	movlw   PB_TRIS		;preset port B direction
        movwf   TRISB
        movlw   OPT_VAL		;preset option register 
        movwf	OPTION_REG
	bcf	STATUS, RP0	;select first bank

;firstly, see if any characters available, and exit if not
	call	CharCount	;check for characters
	bz	hcom_exit	;directly exit if none available

;character available, so read it in and check if zero (already in RxStore)
	call	GetChar		;get character
	bz	hcom_zero	;jump immediately if Z set

;.............................................................................
;non-zero, so check if comms mode 1 (wait header)
	movlw	cm_wait_head	;check against mode 1
	subwf	comms_mode, w	;do check, without affecting comms_mode
	bz	hcom_head	;branch if mode 1

;no match, so check if comms mode 2 (wait code 1)
	movlw	cm_wait_cmd1	;check against mode 2
	subwf	comms_mode, w	;do check, without affecting comms_mode
	bz	hcom_cmd1	;branch if mode 2

;no match, so check if comms mode 3 (wait data)
	movlw	cm_wait_data	;check against mode 3
	subwf	comms_mode, w	;do check, without affecting comms_mode
	bz	hcom_data	;branch if mode 3

;no match, so check if comms mode 4 (wait checksum)
	movlw	cm_wait_csum	;check against mode 4
	subwf	comms_mode, w	;do check, without affecting comms_mode
	bz	hcom_csum	;branch if mode 4

;no match, so check if comms mode 5 (wait term)
	movlw	cm_wait_term	;check against mode 5
	subwf	comms_mode, w	;do check, without affecting comms_mode
	bz	hcom_term	;branch if mode 5

;waiting for zero/invalid comms mode, so reload with zero and exit
	goto	hcom_reset

;.............................................................................
hcom_head
;waiting for header, so check if match with standard header code
	movlw	quicc_idm_hdr	;load header code
	subwf	RxStore, w	;do check, without affecting RxStore
	bnz	hcom_reset	;jump to reset routine if NO MATCH

;header code matches, so select mode 2 (wait code 1)
	movlw	cm_wait_cmd1	;load comms mode
	movwf	comms_mode	;
	goto	hcom_exit	;exit routine

;.............................................................................
hcom_cmd1
;waiting for code 1, so store this in global and checksum
	movfw	RxStore		;get input character
	movwf	cmd_store	;store in global
	movwf	csum_store	;store in checksum as initial value

;change comms mode to waiting for data
	movlw	cm_wait_data	;load comms mode 3 (wait for data)
	movwf	comms_mode	;
	goto	hcom_exit	;exit routine

;.............................................................................
hcom_data
;waiting for data, so store this in global
	movfw	RxStore		;get input character
	movwf	data_store	;store in global

;add to checksum, and replace if it would become zero
	addwf	csum_store, w	;ADD to checksum and store in W
	skpnz			;skip next instr if NOT zero
	movlw	csum_zero_rplc	;replace with this value IF zero
	movwf	csum_store	;store w in checksum

;change comms mode to waiting for checksum
	movlw	cm_wait_csum	;load comms mode 4 (wait for checksum)
	movwf	comms_mode	;
	goto	hcom_exit	;exit routine

;.............................................................................
hcom_csum
;waiting for checksum, so firstly verify against stored checksum
	movfw	RxStore		;get input character
	subwf	csum_store, w	;do check, without affecting csum_store
	bnz	hcom_reset	;jump to reset routine if NO MATCH

;checksums match, so select mode 5 (wait term)
	movlw	cm_wait_term	;load comms mode
	movwf	comms_mode	;
	goto	hcom_exit	;exit routine

;.............................................................................
hcom_term
;waiting for terminator so check if match with standard header code
	movlw	quicc_idm_term	;load terminator code
	subwf	RxStore, w	;do check, without affecting RxStore
	bnz	hcom_reset	;jump to reset routine if NO MATCH

;we have a valid command, so process it
	call	process_cmd

;command executed, so change back to mode 0 (waiting for 0), and exit
	goto	hcom_reset

;.............................................................................
hcom_zero
;zero received, so immediately reset to mode 1 (waiting for startup)
	movlw	cm_wait_head	;change to waiting for header
	movwf	comms_mode	;
	goto	hcom_exit	;exit routine

;.............................................................................
hcom_reset
;reset comms mode to looking for zero
	movlw	cm_wait_zero	;load initial comms mode
	movwf	comms_mode	;

hcom_exit
	return			;directly return

;*****************************************************************************
;				send_status_msg
;
; Sends a "standard issue" status message to the PC, which summarises the
; complete status of the system. This has the structure:
;	
;	1) Header code 0xFE
;	2) Firmware revision
;	3) S_FLAGS status data
;	4) Initial PORT B status (at power-on)
;	5) PORT B status (contains watchdog status)
;	6) Tripped PORT B status (last failure status)
;	7) Boot PORT B ststus (how many panels fitted)
;	8) M_FLAGS status data
;	9) SerialStatus status data
;	10) Standby timer setting
;	11) Boot timer setting
;	12) PSU Override timer setting
;	13) Flash clock 1 dividers
;	14) Flash clock 2 dividers
;	15) dummy value (0x01)
;	16) Checksum value
;	17) 0xFE
;
; Note that no values are ever zero - PORT B ORED with a mask to ensure that
; this does not occur. If checksum would become zero, it is replaced with
; a special value.
;
; NOTE: This routine can be called from within the delay routine, so must
; never call this delay routine!
;
; NOTE2: If config switch 2 is OPEN, this routine will immediately exit,
; as the system is operating in "dumb" mode under these circumstances.
;
;*****************************************************************************

send_status_msg:
;firstly check if in "Dumb mode", config switch 2 OPEN
	btfsc	_cfg_sw2 	;check if config switch 2 is CLOSED
	return			;return directly if OPEN

;transmit a zero seperator
	movlw	0
	call	PutChar

;transmit another zero seperator
	movlw	0
	call	PutChar

;send the standard header code
	movlw	quicc_idm_hdr	;load header code
	call	PutChar

;.............................................................................
;transmit the firmware revision number
	movlw	firmware_ver
	movwf	csum_store	;STORE in checksum as initial value
	call	PutChar		;call character transmitter

;extract S_FLAGS status, OR with mask, and transmit
	movfw	S_FLAGS
	iorlw	stat_sflag_mask
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;extract stored "initial" port B status, OR with mask, and transmit
	movfw	init_portb
	iorlw	stat_portb_mask
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;extract port B status, OR with mask, and transmit
	movfw	PORTB
	iorlw	stat_portb_mask
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;extract stored "tripped" port B status, OR with mask, and transmit
	movfw	trip_portb
	iorlw	stat_portb_mask
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;extract stored "boot" port B status, OR with mask, and transmit
	movfw	boot_portb
	iorlw	stat_portb_mask
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;extract M_FLAGS status, OR with mask, and transmit
	movfw	M_FLAGS
	iorlw	stat_sflag_mask
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;extract Serial status, OR with mask, and transmit
	movfw	SerialStatus
	iorlw	stat_sflag_mask
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;send standby timer value
	movfw	sby_tmr		;get timer value in w
	iorlw	0		;ensure zero flag registers
	skpnz			;check if non zero
	movlw	0ffh		;set to 0ffh if zero
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;send boot timer value
	movfw	boot_tmr	;get timer value in w
	iorlw	0		;ensure zero flag registers
	skpnz			;check if non zero
	movlw	0ffh		;set to 0ffh if zero
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;send PSU override timer value
	movfw	psu_tmr		;get timer value in w
	iorlw	0		;ensure zero flag registers
	skpnz			;check if non zero
	movlw	0ffh		;set to 0ffh if zero
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;send flash 1 setting
	movfw	flash1_set	;get flash setting value in w
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;send flash 2 setting
	movfw	flash2_set	;get flash setting value in w
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;send a dummy byte
	movlw	01h
	addwf	csum_store	;ADD value to checksum store
	call	PutChar		;call character transmitter

;.............................................................................
;now send the accumulated checksum, firstly replacing if zero
	movfw	csum_store	;get stored checksum and set flags
	iorlw	0		;get Z flag from result
	skpnz			;skip next instr if NOT zero
	movlw	csum_zero_rplc	;replace with this value IF zero
	movwf	csum_store	;store for future use
	call	PutChar		;send this character

;finally, send the standard terminator code and exit
	movlw	quicc_idm_term	;load terminator code
	call	PutChar
	return			;directly return


;*****************************************************************************
;				process_cmd
;
; Processes command from PC. Command code is in global cmd_store, and data
; is in global data_store. In all cases, a system ststus message is 
; transmitted once the command is executed or acknowledged. Some commands,
; such as the "wake up" command, action a sequence of events, so this routine
; will send the status message BEFORE this sequence is started.
;
; NOTE: This routine has been implemented as compares and branches, rather 
; than as a call table, because the latter requires interrupts to be disabled.
; In this system, the RTCC interrupt and INT0 interrupt are fairly critical
; as far as timing goes, so they should not be disabled unless absolutely
; necessary.
;
; NOTE: This routine can be called from within the delay routine, so must
; never call this delay routine!
;
; NOTE2: This system start command resets the standby timer. This allows 
; resetting of the standby timer to be performed even if DTR is not toggling.
;
; WARNING! After this routine returns, the character transmission is in
; progress, so nothing must be done directly to the port B output until
; it finished. This applies especially to the standby setting routine.
;
; EEPROM PROTECTION: Due to the fact that random execution may occur on power
; off, potentially corrupting the EEPROM, this routine contains a number of
; safeguards to prevent spurious EEPROM write:
;   
; 1) A seperate EEPROM enable command has to be sent BEFORE any update
; command, thereby preventing EEPROM updates from a single random jump.
; 2) EEPROM address and command codes are nullified once used, thereby
; preventing the interpreter from jumping back into a write routine.
; 3) EEPROM write is disabled once any message has been processed (unless
; it was an enable message). It is also disabled once the update commands
; have been checked for, and on any EEPROM read.
; 4) After setting an EEPROM address, but before doing the write, the command
; code is re-verified. Even if writes have been enabled, this should lock
; out any EEPROM writes to valid addresses, if any part of this module is
; randomly jumped into.
;
;*****************************************************************************

process_cmd:

;check if command matches standby timer set value
	movlw	cmdc_set_sby	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_sby_set	;jump to routine if MATCH

;check if command matches boot timer set value
	movlw	cmdc_set_boot	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_boot_set	;jump to routine if MATCH

;check if command matches PSU timer set value
	movlw	cmdc_set_psu	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_psu_set	;jump to routine if MATCH

;check if command matches default timer setting value
	movlw	cmdc_tmr_dflt	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_timer_dflt	;jump to routine if MATCH

;at this point, disable EEPROM writes, as no further commands use it
	bsf	STATUS, RP0	;select bank 1
	bcf	EECON1, WREN	;disable writes
	bcf	STATUS, RP0	;select bank 0

;check if command matches system wakeup call
	movlw	cmdc_sys_start	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_sys_start	;jump to routine if MATCH

;check if command matches boot hold call
	movlw	cmdc_boot_hold	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_boot_hold	;jump to routine if MATCH

;check if command matches system sleep call
	movlw	cmdc_go_sleep	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_go_sleep	;jump to routine if MATCH

;check if command matches timer eeprom load value
	movlw	cmdc_epm_load	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_eeprom_load	;jump to routine if MATCH

;check if command matches reboot request
	movlw	cmdc_reboot	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_reboot	;jump to routine if MATCH

;check if command matches recover request
	movlw	cmdc_recover	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_recover	;jump to routine if MATCH

;check if command matches display reset request
	movlw	cmdc_dreset	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_dreset	;jump to routine if MATCH

;check if command matches fault lamp test request
	movlw	cmdc_flamp_test	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_flamp_test	;jump to routine if MATCH

;check if command matches flash clock reset
	movlw	cmdc_flash_rst	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_flash_rst	;jump to routine if MATCH

;check if command matches flash clock 1 setting
	movlw	cmdc_flash1_set	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_flash1_set	;jump to routine if MATCH

;check if command matches flash clock 2 setting
	movlw	cmdc_flash2_set	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_flash2_set	;jump to routine if MATCH

;check if command matches flash clock inverse reset
	movlw	cmdc_flash_inv	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_flash_inv	;jump to routine if MATCH

;check if command matches flash clock 1 counter
	movlw	cmdc_flash1_cnt	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_flash1_cnt	;jump to routine if MATCH

;check if command matches flash clock 2 counter
	movlw	cmdc_flash2_cnt	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_flash2_cnt	;jump to routine if MATCH

;check if command matches EEPROM write enable
	movlw	cmdc_enab_eewr	;load command code to check against
	subwf	cmd_store, w	;do check, without affecting cmd_store
	bz	prc_enab	;jump to routine if MATCH

;assume simple status request, so directly jump to exit routine
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_sys_start
;trigger a system start by setting the flag, and reset standby timer
	bsf	_start_request	;set startup request flag
	movfw	sby_tmr		;reset slow standby timer to max
	movwf	CT_SBY		;
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_boot_hold
;trigger a boot hold by setting the flag, and reset standby timer
	bsf	_boot_hold_req	;set boot hold request flag
	movfw	sby_tmr		;reset slow standby timer to max
	movwf	CT_SBY		;
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_go_sleep
;put the system to sleep by setting the flag
	bsf	_sby_request	;ALWAYS set standby request flag
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_sby_set
;update the global standby timer in EEPROM with the new value
	movlw	ee_sby_tmr	;get EEPROM offset
	movwf	EEADR		;store in EEPROM address
	movlw	cmdc_set_sby	;RE-VERIFY command code at this point
	subwf	cmd_store, w	;
	bnz	prc_fin		;jump to exit if NO MATCH
	movfw	data_store	;get stored data in W
	movwf	csum_eeprom	;store INITIAL checksum
	call	write_w_eeprom	;call EEPROM writer with data in W

;calculate EEPROM checksum from existing timer values and this new value
	movfw	boot_tmr	;ADD BOOT timer to checksum
	addwf	csum_eeprom	;
	movfw	psu_tmr		;ADD PSU timer to checksum
	addwf	csum_eeprom	;

;now store this value in the EEPROM
	movlw	ee_checksum	;get EEPROM checksum offset
	movwf	EEADR		;store in EEPROM address
	movlw	cmdc_set_sby	;RE-VERIFY command code at this point
	subwf	cmd_store, w	;
	bnz	prc_fin		;jump to exit if NO MATCH
	movfw	csum_eeprom	;get EEPROM checsum value
	call	write_w_eeprom	;call EEPROM writer with data in W

;reload all global data from EEPROM (also reloads the standby timer itself)
	call	get_global_eeprom_data
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_boot_set
;update the global boot timer with the new value
	movlw	ee_boot_tmr	;get EEPROM offset
	movwf	EEADR		;store in EEPROM address
	movlw	cmdc_set_boot	;RE-VERIFY command code at this point
	subwf	cmd_store, w	;
	bnz	prc_fin		;jump to exit if NO MATCH
	movfw	data_store	;get stored data in W
	movwf	csum_eeprom	;store INITIAL checksum
	call	write_w_eeprom	;call EEPROM writer with data in W

;calculate EEPROM checksum from existing timer values and this new value
	movfw	sby_tmr		;ADD standby timer to checksum
	addwf	csum_eeprom	;
	movfw	psu_tmr		;ADD PSU timer to checksum
	addwf	csum_eeprom	;

;now store this value in the EEPROM
	movlw	ee_checksum	;get EEPROM checksum offset
	movwf	EEADR		;store in EEPROM address
	movlw	cmdc_set_boot	;RE-VERIFY command code at this point
	subwf	cmd_store, w	;
	bnz	prc_fin		;jump to exit if NO MATCH
	movfw	csum_eeprom	;get EEPROM checsum value
	call	write_w_eeprom	;call EEPROM writer with data in W

;reload all global data from EEPROM, disabling writes in process
	call	get_global_eeprom_data
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_psu_set
;update the global psu override timer with the new value
	movlw	ee_psu_tmr	;get EEPROM offset
	movwf	EEADR		;store in EEPROM address
	movlw	cmdc_set_psu	;RE-VERIFY command code at this point
	subwf	cmd_store, w	;
	bnz	prc_fin		;jump to exit if NO MATCH
	movfw	data_store	;get stored data in W
	movwf	csum_eeprom	;store INITIAL checksum
	call	write_w_eeprom	;call EEPROM writer with data in W

;calculate EEPROM checksum from existing timer values and this new value
	movfw	sby_tmr		;ADD standby timer to checksum
	addwf	csum_eeprom	;
	movfw	boot_tmr	;ADD BOOT timer to checksum
	addwf	csum_eeprom	;

;now store this value in the EEPROM
	movlw	ee_checksum	;get EEPROM checksum offset
	movwf	EEADR		;store in EEPROM address
	movlw	cmdc_set_psu	;RE-VERIFY command code at this point
	subwf	cmd_store, w	;
	bnz	prc_fin		;jump to exit if NO MATCH
	movfw	csum_eeprom	;get EEPROM checsum value
	call	write_w_eeprom	;call EEPROM writer with data in W

;reload all global data from EEPROM, disabling writes in process
	call	get_global_eeprom_data
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_timer_dflt
;store default standby timer value
	movlw	ee_sby_tmr	;get EEPROM standby timer offset
	movwf	EEADR		;store in EEPROM address
	movlw	cmdc_tmr_dflt	;RE-VERIFY command code at this point
	subwf	cmd_store, w	;
	bnz	prc_fin		;jump to exit if NO MATCH
	movlw	TS_SBY		;load default standby timer value
	call	write_w_eeprom	;call EEPROM writer with data in W

;store default boot timer value
	movlw	ee_boot_tmr	;get EEPROM boot timer offset
	movwf	EEADR		;store in EEPROM address
	movlw	cmdc_tmr_dflt	;RE-VERIFY command code at this point
	subwf	cmd_store, w	;
	bnz	prc_fin		;jump to exit if NO MATCH
	movlw	TM_BOOT		;load default boot timer value
	call	write_w_eeprom	;call EEPROM writer with data in W

;store default psu override timer value
	movlw	ee_psu_tmr	;get EEPROM psu timer offset
	movwf	EEADR		;store in EEPROM address
	movlw	cmdc_tmr_dflt	;RE-VERIFY command code at this point
	subwf	cmd_store, w	;
	bnz	prc_fin		;jump to exit if NO MATCH
	movlw	TM_PSU		;load default PSU timer value
	call	write_w_eeprom	;call EEPROM writer with data in W

;update EEPROM checksum with pre-calculated value
	movlw	ee_checksum	;get EEPROM offset
	movwf	EEADR		;store in EEPROM address
	movlw	cmdc_tmr_dflt	;RE-VERIFY command code at this point
	subwf	cmd_store, w	;
	bnz	prc_fin		;jump to exit if NO MATCH
	movlw	EE_CHECK	;load default checksum value
	call	write_w_eeprom	;call EEPROM writer with data in W

;reload all global data from EEPROM, disabling writes in process
	call	get_global_eeprom_data
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_eeprom_load
;reload all timers from EEPROM memory (unless config switch 3 open)
	call	get_global_eeprom_data
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_reboot
;switch off PSU and reset displays, also illuminating boot LED
	bcf	PORTA, PSU_ON	;switch off the PSU
	bsf	_pic_wdog_boot	;indicate watchdog re-boot in progress
	bcf	PORTA, DP_RST	;put the display panels in reset
	bcf	PORTB, BOOT_LD	;illuminate the "boot" LED
	bsf	PORTA, FAIL_LD	;extinguish the "fault" LED

;now directly send back status and then just loop, allowing watchdog trip
	call	send_status_msg
prc_loop1
	goto	prc_loop1	;loop to allow watchdog trip

;.............................................................................
prc_recover
;recover from fault mode, by setting flag
	bsf	_recover_req	;simply set recovery request flag
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_dreset
;request display reset, by setting flag
	bsf	_dreset_req	;simply set display reset request flag
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_flamp_test
;request lamp test, by setting flag
	bsf	_flamp_test_req	;simply set fault lamp test request flag
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_flash_rst
;simply reset both flash clocks directly, and load counters
	bcf	PORTA, FLASH1	;CLEAR flash clock 1
	movfw	flash1_set	;load set value in W
	movwf	flash1_cnt	;put this into counter
	bcf	PORTA, FLASH2	;CLEAR flash clock 2
	movfw	flash2_set	;load set value in W
	movwf	flash2_cnt	;put this into counter
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_flash1_set
;update the global flash 1 setter the new value
	movfw	data_store	;copy data store to flash 1 set
	movwf	flash1_set	;
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_flash2_set
;update the global flash 2 setter the new value
	movfw	data_store	;copy data store to flash 2 set
	movwf	flash2_set	;
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_flash_inv
;simply inverse reset both flash clocks, and load counters
	bcf	PORTA, FLASH1	;CLEAR flash clock 1
	movfw	flash1_set	;load set value in W
	movwf	flash1_cnt	;put this into counter
	bsf	PORTA, FLASH2	;SET flash clock 2
	movfw	flash2_set	;load set value in W
	movwf	flash2_cnt	;put this into counter
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_flash1_cnt
;update the global flash 1 counter the new value
	movfw	data_store	;copy data store to flash 1 count
	movwf	flash1_cnt	;
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_flash2_cnt
;update the global flash 2 counter the new value
	movfw	data_store	;copy data store to flash 2 count
	movwf	flash2_cnt	;
	goto	prc_fin		;jump to exit routine

;.............................................................................
prc_fin
;finished processing command so disable EEPROM writes and send status
	bsf	STATUS, RP0	;select bank 1
	bcf	EECON1, WREN	;disable writes
	bcf	STATUS, RP0	;select bank 0
	call	send_status_msg	;send status message
	movlw	cmdc_get_stat	;nullify command store
	movwf	cmd_store	;
	return			;directly return

;.............................................................................
prc_enab
;finished processing command, so send status and enable EEPROM writes
	call	send_status_msg	;send status message
	bsf	STATUS, RP0	;select bank 1
	bsf	EECON1, WREN	;enable writes
	bcf	STATUS, RP0	;select bank 0
	movlw	cmdc_get_stat	;nullify command store
	movwf	cmd_store	;
	return			;directly return

;*****************************************************************************
;********************  EEPROM DATA HANDLING SUBROUTINES  *********************
;*****************************************************************************

;*****************************************************************************
;				read_eeprom
;
; Reads data from EEPROM, using location specified by W register. Data value
; is returned in W register, and is also available from EEDATA location.
;
; NOTE: To help prevent spurious writes, this routine sets the EEPROM
; address to a null value once the read has complated.
;
;*****************************************************************************

read_eeprom
	movwf	EEADR		;store W register in EEPROM address
	bsf	STATUS, RP0	;select bank 1
	bcf	EECON1, WREN	;disable writes
	bsf	EECON1, RD	;initiate read cycle
	bcf	STATUS, RP0	;select bank 0
	movlw	ee_null_addr	;set to null EEPROM address
	movwf	EEADR		;
	movfw	EEDATA		;store data in W
	return

;*****************************************************************************
;				write_eeprom
;
; Writes data to EEPROM, using data and address already loaded in 
; EEDATA and EEADR, respectively. Note that interrupts are disabled for
; part of this routines, so serial comms should NOT be in progress when
; this routine is called.
;
; NOTE: If the WRERR bit in EECON1 is set on startup, the PIC was reset 
; before the write had a chance to complete. This routine can safely be 
; called to re-initiate the write as EEDATA and EEADR will contain the
; original values.
;
; After calling this routine, the caller should poll the EEIF flag in the
; EECON1 register, which will be set once the write has completed.
;
;*****************************************************************************

write_eeprom
;firstly check if EEPROM disabled, config switch 3 OPEN
	btfsc	_cfg_sw3 	;check if config switch 3 is CLOSED
	return			;return directly if OPEN

;select bank 1 ready for interrupt and EEPROM control
	bsf	STATUS, RP0	;select bank 1

;now disable interrupts, and ensure they are disabled
wrep_loop1
	bcf	INTCON, GIE	;disable global ints
	btfsc	INTCON, GIE	;check that they are disabled
	goto	wrep_loop1	;loop round if still enabled

;now enable EEPROM writes and issue EXACT write sequence
	movlw	055h		;write 55h to EECON2
	movwf	EECON2
	movlw	0aah		;write aah to EECON2
	movwf	EECON2
	bsf	EECON1, WR	;initiate write

;enable ints, select bank 0 and return
	bsf	INTCON, GIE	;enable global ints
	bcf	STATUS, RP0	;select bank 0
	return

;*****************************************************************************
;				write_w_eeprom
;
; High level routine checks that config switch 3 is closed (EEPROM enable),
; and that the data to be written in W is different from that already stored.
; If so, writes W register into EEPROM data area. EEPROM address must still
; have been specified beforehand. After initiating the write, the routine
; WAITS until the write has completed. Note that NO timing or comms handlers
; can be called, because this is called from within them!
;
; NOTE: This routine relies on the watchdog timer to reset the system if
; for some reason the write doesn't complete. However, the timer MUST have 
; the prescaler enabled to a high value (timeout half a second or so). Using
; the watchdog timer without a prescaler will cause problems, as it could
; theoretically be less that the EEPROM write time.
;
; NOTE2: To help prevent spurious writes, this routine sets the EEPROM
; address to a null value once the write has complated.
;
;*****************************************************************************

write_w_eeprom
;firstly check if EEPROM disabled, config switch 3 OPEN
	btfsc	_cfg_sw3 	;check if config switch 3 is CLOSED
	return			;return directly if OPEN

;now check if EEPROM write enable flag clear
	bsf	STATUS, RP0	;select bank 1
	btfsc	EECON1, WREN	;check if write enable CLEAR
	goto	wwep_cont1	;continue execution if SET
	bcf	STATUS, RP0	;select bank 0
	return			;return directly if OPEN

wwep_cont1
;check if EEPROM already contains specified data
	bsf	EECON1, RD	;initiate read cycle (already in bank 1)
	bcf	STATUS, RP0	;select first bank
	subwf	EEDATA, F	;subtract W from data, PRESERVING W
	bnz	wwep_write	;jump to write if no match
	return			;return if match

wwep_write
;now transfer W register into EEPROM data and call write routine
	movwf	EEDATA		;store W register in EEPROM data
	call	write_eeprom	;initiate EEPROM write

;now WAIT until EEPROM write has completed
	clrwdt			;clear down watchdog timer
	bsf	STATUS, RP0	;select second bank
wwep_wait1
	btfsc	EECON1, EEIF	;test for EEPROM write complete CLEAR
	goto	wwep_fin1	;exit loop if flag has SET
	goto	wwep_wait1	;loop round for another go

;flag has been triggered, so clear it down and return
wwep_fin1	
	bcf	EECON1, EEIF	;clear EEPROM write complete flag
	bcf	STATUS, RP0	;select first bank
	movlw	ee_null_addr	;set to null EEPROM address
	movwf	EEADR		;
	return			;return to caller

;*****************************************************************************
;			get_global_eeprom_data
;
; High level routine which extracts all global data from EEPROM. Also checks
; that EEPROM checksum is correct, and controls checksum flag appropriately.
; Note that this routine can also be used to STORE a new checksum, as on
; exit, csum_eeprom will contain the new calculated checksum.
;
; NOTE: This routine also updates the actual standby timer as well as the
; store, so that the next timeout will reflect the updated value.
;
;*****************************************************************************

get_global_eeprom_data
;firstly, clear down the bad checksum flag
	bcf	_bad_ee_check	;indicate good checksum

;extract and store standby timer setting, and update standby timer
	movlw	ee_sby_tmr	;get EEPROM offset
	call	read_eeprom	;get EEPROM data in W
	movwf	sby_tmr		;store in live global
	movwf	CT_SBY		;update actual timer value
	movwf	csum_eeprom	;store INITIAL checksum value

;extract and store boot timer setting
	movlw	ee_boot_tmr	;get EEPROM offset
	call	read_eeprom	;get EEPROM data in W
	movwf	boot_tmr	;store in live global
	addwf	csum_eeprom	;ADD value to checksum store

;extract and store psu override timer setting
	movlw	ee_psu_tmr	;get EEPROM offset
	call	read_eeprom	;get EEPROM data in W
	movwf	psu_tmr		;store in live global
	addwf	csum_eeprom	;ADD value to checksum store

;extract stored checksum, and check against calculated value
	movlw	ee_checksum	;get EEPROM offset
	call	read_eeprom	;get EEPROM data in W
	subwf	csum_eeprom, w	;do check, without affecting csum_eeprom
	skpz			;skip next instruction if MATCH
	bsf	_bad_ee_check	;indicate BAD checksum if NO match

;return to caller (csum_eeprom contains calculated checksum)
	return

;-----------------------------------------------------------------------------
	END

;*****************************************************************************
;********** End of File ******** End of File ******** End of File ************
